#include <sys/io.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <libx86.h>
#include <netinet/in.h>
#include "vesamode.h"
#include "edid.h"


void fatal(char *format, ...) __attribute__ ((format (printf, 1, 2)));
void fatal(char *format, ...){
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    exit(1);
}
void warning(char *format, ...) __attribute__ ((format (printf, 1, 2)));
void warning(char *format, ...){
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
}

/* Create a 'canonical' version of a string,
 * i.e. no spaces at start and end.
 */
static char *canon_str(char *s){
  char *start, *end;

  int len = strlen(s);
  for (start = s; start < s + len; start++) if (!isspace(*start)) break;

  for (end = s+len-1; end >= start; --end) {
    if (!isspace(*end)) break;
    *end = '\0';
  }
  return start;
}

void *segofs2addr(uint8_t segofs[4]){
  uintptr_t ret=segofs[0] + (segofs[1] << 8) + (segofs[2] << 4)+ (segofs[3] << 12);
  return (void *)ret;
}

// get string pointed by str_ptr address with range checking
char *get_string(char *OemData, uint8_t str_ptr[4]){
  // String must be either in 0xC0000-0x100000 range (values comes from the original ddcprobe),
  // or in OemData[256] block. Recommended string size is < 256b
  char buf[1024], *p=(char *)segofs2addr(str_ptr);

  unsigned int i;
  for (i=0; i<sizeof(buf); i++){
    if (((p<OemData) || (p>=OemData+256)) &&
        ((p<(char *)0xC0000) || (p>=(char *)0x100000))){
      warning("VBE: warning: range check error for string\n");
      break;
    }
    if (*p == '\0') break;
    buf[i]=*p; p++;
  }
  buf[i]='\0';
  return strdup(canon_str(buf));
}

// do range checking, get pointer to modes array
uint16_t *get_modes(char *ReservedData, uint8_t str_ptr[4]){
  // Modes must be either in 0xC0000-0x100000 range (values comes from the original ddcprobe),
  // or in ReservedData[222] block.
  // Each mode number occupies one word (16 bits).
  // The list of mode numbers is terminated by a -1 (0FFFFh).
  uint16_t *ret=segofs2addr(str_ptr), *p;

  for (p=ret; *p!=0xFFFF; p++){
    if (((p<(uint16_t *)ReservedData) || (p>=(uint16_t *)(ReservedData+222))) &&
        ((p<(uint16_t *)0xC0000) || (p>=(uint16_t *)0x100000))){
      warning("VBE: warning: range check error for modes\n");
      return NULL;
    }
  }
  return ret;
}

char *snip(char *string){
  int i;
  /* This is always a 13 character buffer */
  /* and it's not always terminated. */
  string[12] = '\0';

  while(((i = strlen(string)) > 0) &&
         (isspace(string[i - 1]) ||
          (string[i - 1] == '\n') ||
          (string[i - 1] == '\r'))) {
    string[i - 1] = '\0';
  }
  return string;
}


/*********************************************************************/
void print_vbe_info(){ // Print VBE Controller Information
  /* See VESA BIOS Extensions (VBE) Core Functions Standard, v.3.0, p.25 */

  int i,j,k, s;
  struct LRMI_regs regs;

  struct VbeInfoBlock_t{
    char     VbeSignature[4];      // VBE Signature, "VESA"
    uint16_t VbeVersion;           // VBE Version, 0x0300
    uint8_t  OemStringPtr[4];      // VbeFarPtr to OEM String
    uint32_t Capabilities;         // Capabilities of graphics controller
    uint8_t  VideoModePtr[4];      // VbeFarPtr to VideoModeList
    uint16_t TotalMemory;          // Number of 64kb memory blocks, Added for VBE 2.0+
    uint16_t OemSoftwareRev;       // VBE implementation Software revision
    uint8_t  OemVendorNamePtr[4];  // VbeFarPtr to Vendor Name String
    uint8_t  OemProductNamePtr[4]; // VbeFarPtr to Product Name String
    uint8_t  OemProductRevPtr[4];  // VbeFarPtr to Product Revision String
    char     Reserved[222];        // Reserved for VBE implementation scratch area
    char     OemData[256];         // Data Area for OEM Strings
  } __attribute__((packed)) * VbeInfoBlock;

  // Getting VbeInfoBlock!
  s=sizeof(struct VbeInfoBlock_t);
  if (s != 512) fatal("VBE: VbeInfoBlock struct has wrong size: %d\n", s);

  VbeInfoBlock = (struct VbeInfoBlock_t *)LRMI_alloc_real(s);
  if (VbeInfoBlock == NULL) fatal("VBE: could allocate real memory\n");
  if (((uintptr_t)VbeInfoBlock) >> 20) fatal("VBE: invalid LRMI pointer\n");

  memset(VbeInfoBlock, 0, s);

  // set up registers for the interrupt call
  // Int 10, function 0x00 -- Return VBE Controller Information.
  // On input AX=0x4F00, ES:DI - pointer to 512-byte buffer.
  // first 4 bytes should be set to "VBE2"!
  memset(&regs, 0, sizeof(regs));
  regs.eax = 0x4f00;
  regs.es = ((uintptr_t)VbeInfoBlock) >> 4;
  regs.edi = ((uintptr_t)VbeInfoBlock) & 0x0F;
  memcpy(VbeInfoBlock->VbeSignature, "VBE2", 4);

  // do it
  iopl(3);
  ioperm(0, 0x400, 1);
  if (LRMI_int(0x10, &regs) == 0) {
    LRMI_free_real(VbeInfoBlock);
    fatal("VBE: Error on LRMI_int\n");
  }

  // check for successful return code
  i = regs.eax & 0xffff;
  if (i != 0x4f) {
    LRMI_free_real(VbeInfoBlock);
    fatal("VBE: Error (0x4f00): 0x%04x\n", i);
  }

  // check for "VESA" string
  if (memcmp(VbeInfoBlock->VbeSignature, "VESA", 4)){
    LRMI_free_real(VbeInfoBlock);
    fatal("VBE: Error in VbeSignature\n");
  }

  printf("vbe: %c%c%c%c %d.%d\n",
               VbeInfoBlock->VbeSignature[0],
               VbeInfoBlock->VbeSignature[1],
               VbeInfoBlock->VbeSignature[2],
               VbeInfoBlock->VbeSignature[3],
               VbeInfoBlock->VbeVersion >>8,
               VbeInfoBlock->VbeVersion &0xFF);

  printf("oem: %s\n",
    get_string(VbeInfoBlock->OemData, VbeInfoBlock->OemStringPtr));
  if ((VbeInfoBlock->VbeVersion >> 8) >= 3){
    printf("vendor: %s\n",
      get_string(VbeInfoBlock->OemData, VbeInfoBlock->OemVendorNamePtr));
    printf("product: %s %s\n",
      get_string(VbeInfoBlock->OemData, VbeInfoBlock->OemProductNamePtr),
      get_string(VbeInfoBlock->OemData, VbeInfoBlock->OemProductRevPtr));
  }


  printf("memory: %dkb\n", VbeInfoBlock->TotalMemory << 6);

  uint16_t *modes=get_modes(VbeInfoBlock->Reserved, VbeInfoBlock->VideoModePtr);

  // List supported standard modes
  for (i=0; i<4; i++) {
    switch (i) {
      case 0: printf("8bit:");  break;
      case 1: printf("16bit:"); break;
      case 2: printf("24bit:"); break;
      case 3: printf("32bit:"); break;
    }
    if (modes != NULL){
      for(j=0; modes[j] != 0xffff; j++) {
        for(k = 0; known_vesa_modes[k].x != 0; k++) {
          if (known_vesa_modes[k].number == modes[j]) {
            u_int32_t colors = known_vesa_modes[k].colors-1;
            if ((colors == 0x0000ff && i == 0) ||
                (colors == 0x007fff && i == 1) ||
                (colors == 0x00ffff && i == 2) ||
                (colors == 0xffffff && i == 3)) {
              printf(" %dx%d", known_vesa_modes[k].x, known_vesa_modes[k].y);
            }
          }
        }
      }
    }
    printf("\n");
  }

  LRMI_free_real(VbeInfoBlock);
}

/*********************************************************************/
void print_vbe_mode(){ // Print current VBE mode
  /* See VESA BIOS Extensions (VBE) Core Functions Standard, v.3.0, p.25 */

  int i;
  struct LRMI_regs regs;

  // set up registers for the interrupt call
  // Int 10, function 0x03 -- Return Current VBE Mode
  // On input AX=0x4F03
  memset(&regs, 0, sizeof(regs));
  regs.eax = 0x4f03;

  // do it
  iopl(3);
  ioperm(0, 0x400, 1);
  if (LRMI_int(0x10, &regs) == 0) fatal("VBE: Error on LRMI_int\n");

  // check for successful return code
  i = regs.eax & 0xffff;
  if (i != 0x4f) fatal("VBE: Error (0x4f03): 0x%04x\n", i);

  int mode=regs.ebx & 0x1FFF;

  for(i = 0; known_vesa_modes[i].x != 0; i++) {
    if (known_vesa_modes[i].number == mode) {
      printf("current_mode: 0x%X %dx%d\n", mode, known_vesa_modes[i].x, known_vesa_modes[i].y);
      printf("current_colors: %d\n", known_vesa_modes[i].colors-1);
    }
  }
}

/*********************************************************************/
void print_vbe_ddc_cap(){ // Print VBE/DDC capabilities
  /* See VESA BIOS Extensions/Display Data Channel Standard, v.1.1 */

  int i;
  struct LRMI_regs regs;

  // set up registers for the interrupt call
  memset(&regs, 0, sizeof(regs));
  regs.eax = 0x4f15;
  regs.ebx = 0;
  regs.es  = 0;
  regs.edi = 0;

  // do it
  iopl(3);
  ioperm(0, 0x400, 1);
  if (LRMI_int(0x10, &regs) == 0) fatal("VBE: Error on LRMI_int\n");

  // check for successful return code
  i = regs.eax & 0xffff;
  if (i != 0x4f) fatal("VBE: Error (0x4f15): 0x%04x\n", i);

  printf("ddc_transfer_time: %d s\n",
    (regs.ebx >> 4) && 0xF);
  printf("ddc_support:%s%s\n",
    regs.ebx&1 ? " DDC1": "",
    regs.ebx&2 ? " DDC2": "");
  printf("ddc_screen_blanked: %s\n",
    regs.ebx&3 ? "yes": "no");
}



/*********************************************************************/
void print_vbe_edid(){ // Print VBE Controller Information
  /* See VESA BIOS Extensions/Display Data Channel Standard, v.1.1 
     for int10 function and
     http://www.indopedia.org/Extended_display_identification_data.html
     for edid data structure */
  int i;
  struct LRMI_regs regs;

  uint8_t *edid;
  const uint8_t edid_v1_header[] = { 0x00, 0xff, 0xff, 0xff,
                                     0xff, 0xff, 0xff, 0x00 };
  const int edid_size=128;

  edid = (uint8_t *)LRMI_alloc_real(edid_size);
  if (edid == NULL) fatal("VBE: could allocate real memory\n");
  if (((uintptr_t)edid) >> 20) fatal("VBE: invalid LRMI pointer\n");

  memset(edid, 0, edid_size);

  // set up registers for the interrupt call
  memset(&regs, 0, sizeof(regs));
  regs.eax = 0x4f15; // VESA Extension // VBE/DDC Services
  regs.ebx = 0x0001; // Read EDID
  regs.ecx = 0x0000; // Controller unit number (00 = primary controller)
  regs.edx = 0x0000; // EDID block numder
  regs.es  = ((uintptr_t)edid) >> 4;
  regs.edi = ((uintptr_t)edid) & 0x0F;
    // ES:DI -- Pointer to area in which the EDID block (128 bytes) shall be returned

  // do it
  iopl(3);
  ioperm(0, 0x400, 1);
  if (LRMI_int(0x10, &regs) == 0){
    LRMI_free_real(edid);
    fatal("VBE: Error on LRMI_int\n");
  }

  // check for successful return code
  i = regs.eax & 0xffff;
  if (i != 0x4f){
    LRMI_free_real(edid);
    fatal("VBE: Error (0x4f15): 0x%04x\n", i);
  }

  if (memcmp(edid, edid_v1_header, sizeof(edid_v1_header)))
    warning("VBE: warning: bad EDIDv1 header\n");

  uint8_t chksum=0;
  for (i=0; i<edid_size; i++) chksum+=edid[i];

  if (chksum != 0)
    warning("VBE: warning: bad EDID checksum\n");

  struct edid_info_t edid_struct;
  if (sizeof(edid_struct) != edid_size)
    fatal("VBE: EDID struct has wrong size: %zd\n", sizeof(edid_struct));

  memcpy(&edid_struct, edid, edid_size);
  LRMI_free_real(edid);

  uint32_t ver=(edid_struct.version << 16) + edid_struct.revision;
  printf("edid: %d %d\n",
         edid_struct.version, edid_struct.revision);

  uint16_t man = edid_struct.manufacturer_name[1] +
                (edid_struct.manufacturer_name[0] << 8);

  char manufacturer[4];
  manufacturer[0] = ((man >> 10) & 0x1F) + 'A' - 1;
  manufacturer[1] = ((man >> 5)  & 0x1F) + 'A' - 1;
  manufacturer[2] = (man         & 0x1F) + 'A' - 1;
  manufacturer[3] = '\0';

  uint16_t code=edid_struct.product_code[0] +
               (edid_struct.product_code[1] << 8);

  printf("id: %04x\n", code);
  printf("eisa: %s%04x\n", manufacturer, code);

  if(edid_struct.serial_number != 0xffffffff) {
    if(strcmp(manufacturer, "MAG") == 0) {
      edid_struct.serial_number -= 0x7000000;
    }
    if(strcmp(manufacturer, "OQI") == 0) {
      edid_struct.serial_number -= 456150000;
    }
    if(strcmp(manufacturer, "VSC") == 0) {
      edid_struct.serial_number -= 640000000;
    }
  }
  printf("serial: %08x\n", edid_struct.serial_number);

  printf("manufacture: %d %d\n",
    edid_struct.week, edid_struct.year + 1990);

  printf("input: %s%s%s%s.\n",
    edid_struct.video_input_definition.separate_sync ?
      "separate sync, " : "",
    edid_struct.video_input_definition.composite_sync ?
      "composite sync, " : "",
    edid_struct.video_input_definition.sync_on_green ?
      "sync on green, " : "",
    edid_struct.video_input_definition.digital ?
      "digital signal" : "analog signal");

  printf("screensize: %d %d\n",
    edid_struct.max_size_horizontal,
    edid_struct.max_size_vertical);

  printf("gamma: %f\n", edid_struct.gamma / 100.0 + 1);

  printf("dpms: %s, %s%s, %s%s, %s%s\n",
    edid_struct.feature_support.rgb ? "RGB" : "non-RGB",
    edid_struct.feature_support.active_off ? "" : "no ", "active off",
    edid_struct.feature_support.suspend ? "" : "no ", "suspend",
    edid_struct.feature_support.standby ? "" : "no ", "standby");

  if(edid_struct.established_timings.timing_720x400_70)
    printf("timing: 720x400@70 Hz (VGA 640x400, IBM)\n");
  if(edid_struct.established_timings.timing_720x400_88)
    printf("timing: 720x400@88 Hz (XGA2)\n");
  if(edid_struct.established_timings.timing_640x480_60)
    printf("timing: 640x480@60 Hz (VGA)\n");
  if(edid_struct.established_timings.timing_640x480_67)
    printf("timing: 640x480@67 Hz (Mac II, Apple)\n");
  if(edid_struct.established_timings.timing_640x480_72)
    printf("timing: 640x480@72 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_640x480_75)
    printf("timing: 640x480@75 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_800x600_56)
    printf("timing: 800x600@56 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_800x600_60)
    printf("timing: 800x600@60 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_800x600_72)
    printf("timing: 800x600@72 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_800x600_75)
    printf("timing: 800x600@75 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_832x624_75)
    printf("timing: 832x624@75 Hz (Mac II)\n");
  if(edid_struct.established_timings.timing_1024x768_87i)
    printf("timing: 1024x768@87 Hz Interlaced (8514A)\n");
  if(edid_struct.established_timings.timing_1024x768_60)
    printf("timing: 1024x768@60 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_1024x768_70)
    printf("timing: 1024x768@70 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_1024x768_75)
    printf("timing: 1024x768@75 Hz (VESA)\n");
  if(edid_struct.established_timings.timing_1280x1024_75)
    printf("timing: 1280x1024@75 (VESA)\n");

  /* Standard timings.
  xresolution: Horizontal resolution.  Multiply by 8, then add 248 for actual value.
  aspect:      Aspect ratio. Actual vertical resolution depends on horizontal resolution.
               00=16:10, 01=4:3, 10=5:4, 11=16:9 (00=1:1 prior to v1.3)
  vfreq:       Vertical frequency. Add 60 to get actual value.
  */
  for(i = 0; i < 8; i++) {
    uint16_t xres = edid_struct.standard_timing[i].xresolution;
    uint16_t vfreq = edid_struct.standard_timing[i].vfreq;
    uint16_t x = xres*8 + 248;
    uint16_t y=x;

    if (((xres != 0) && (xres != 1)) ||
        ((vfreq != 0) && (vfreq != 1))) {
      switch(edid_struct.standard_timing[i].aspect) {
        case 0:
          if (ver >= (1 << 16) + 3) // v>=1.3
            y=(10*x)/16;
          else
            y=x;
        break;
        case 1: y=(3*x)/4; break;
        case 2: y=(4*x)/5; break;
        case 3: y=(9*x)/16; break;
      }
      printf("ctiming: %dx%d@%d\n", x, y, vfreq +60);
    }
  }

  /* Detailed timing information. */
  /* The original dtiming code didn't actually work at all, so I've 
   * removed it and replaced it with my own dtiming code, which is derived 
   * from the VESA spec and parse-edid.c. How well it works on monitors 
   * with multiple dtimings is unknown, since I don't have one. -daniels */
  struct edid_monitor_descriptor *monitor = NULL;
  struct edid_detailed_timing *timing;
  for(i = 0; i < 4; i++) {
    timing = &edid_struct.monitor_details.detailed_timing[i];
    if (!timing->pixel_clock) {
      monitor = &edid_struct.monitor_details.monitor_descriptor[i];
      if (monitor->type == edid_monitor_descriptor_serial)
        printf("monitorserial: %s\n", snip(monitor->data.string));
      else if (monitor->type == edid_monitor_descriptor_ascii)
        printf("monitorid: %s\n", snip(monitor->data.string));
      else if (monitor->type == edid_monitor_descriptor_name)
        printf("monitorname: %s\n", snip(monitor->data.string));
      else if (monitor->type == edid_monitor_descriptio_mf_range) {
        printf("monitor_mf_range\n");
      }
      else if (monitor->type == edid_monitor_descriptor_range) {
        // TODO: get more code from monitor-parse-edid
        printf("horizsync: %d-%d\nvertrefresh: %d-%d\n",
               monitor->data.range_data.horizontal_min,
               monitor->data.range_data.horizontal_max,
               monitor->data.range_data.vertical_min,
               monitor->data.range_data.vertical_max);
      }
    }
    else {
      int h_active, h_blanking;
      int v_active, v_blanking;
      int hsync_offset, hsync_pulse_width;
      int vsync_offset, vsync_pulse_width;
      int pixclock;
      int h_total, v_total, vfreq;
      pixclock = EDID_DETAILED_TIMING_PIXEL_CLOCK(*timing);
      h_blanking = EDID_DETAILED_TIMING_H_BLANKING(*timing);
      v_blanking = EDID_DETAILED_TIMING_V_BLANKING(*timing);;
      h_active = EDID_DETAILED_TIMING_H_ACTIVE(*timing);
      v_active = EDID_DETAILED_TIMING_V_ACTIVE(*timing);
      h_total = h_active + h_blanking;
      v_total = v_active + v_blanking;
      hsync_offset = EDID_DETAILED_TIMING_HSYNC_OFFSET(*timing);
      vsync_offset = EDID_DETAILED_TIMING_VSYNC_OFFSET(*timing);
      hsync_pulse_width = EDID_DETAILED_TIMING_HSYNC_PULSE_WIDTH(*timing);
      vsync_pulse_width = EDID_DETAILED_TIMING_VSYNC_PULSE_WIDTH(*timing);
      vfreq = (double)pixclock/((double)v_total*(double)h_total);
      printf("modeline: \"%dx%d@%d\" %.2f %d %d %d %d %d %d %d %d %chsync %cvsync\n",
        h_active, v_active, vfreq, (double)(pixclock/1000000.0),
        h_active, h_active+hsync_offset, h_active+hsync_offset+hsync_pulse_width,
        h_total,
        v_active, v_active+vsync_offset, v_active+vsync_offset+vsync_pulse_width,
        v_total,
        ((timing->flags.variant & 2) == 2 ? '+' : '-'),
        ((timing->flags.variant & 1) == 1 ? '+' : '-'));
    }
  }

}

/*********************************************************************/
int main(){

  if (geteuid())
    fatal("VBE: Permission denied\n");
  if (LRMI_init() == 0)
    fatal("VBE: could not initialize LRMI\n");

  print_vbe_info();
//print_vbe_mode();
//print_vbe_ddc_cap();
  print_vbe_edid();
  return 0;
}
