#ifndef EDID_H
#define EDID_H

#include <stdint.h>

/* Aspect ratios used in EDID info. */
enum edid_aspect {
	aspect_unknown = 0,
	aspect_75,
	aspect_8,
	aspect_5625,
};

/* Detailed timing information used in EDID v1.x */
struct edid_detailed_timing {
	u_int16_t pixel_clock;
#define EDID_DETAILED_TIMING_PIXEL_CLOCK(_x) \
	((_x).pixel_clock * 10000)
	unsigned char h_active;
	unsigned char h_blanking;
	unsigned char h_blanking_hi: 4;
	unsigned char h_active_hi: 4;
#define EDID_DETAILED_TIMING_H_ACTIVE(_x) \
	(((_x).h_active_hi << 8) + (_x).h_active)
#define EDID_DETAILED_TIMING_H_BLANKING(_x) \
	(((_x).h_blanking_hi << 8) + (_x).h_blanking)
	unsigned char v_active;
	unsigned char v_blanking;
	unsigned char v_blanking_hi: 4;
	unsigned char v_active_hi: 4;
#define EDID_DETAILED_TIMING_V_ACTIVE(_x) \
	(((_x).v_active_hi << 8) + (_x).v_active)
#define EDID_DETAILED_TIMING_V_BLANKING(_x) \
	(((_x).v_blanking_hi << 8) + (_x).v_blanking)
	unsigned char hsync_offset;
	unsigned char hsync_pulse_width;
	unsigned char vsync_offset: 4;
	unsigned char vsync_pulse_width: 4;
	unsigned char hsync_offset_hi: 2;
	unsigned char hsync_pulse_width_hi: 2;
	unsigned char vsync_offset_hi: 2;
	unsigned char vsync_pulse_width_hi: 2;
#define EDID_DETAILED_TIMING_HSYNC_OFFSET(_x) \
	(((_x).hsync_offset_hi << 8) + (_x).hsync_offset)
#define EDID_DETAILED_TIMING_HSYNC_PULSE_WIDTH(_x) \
	(((_x).hsync_pulse_width_hi << 8) + (_x).hsync_pulse_width)
#define EDID_DETAILED_TIMING_VSYNC_OFFSET(_x) \
	(((_x).vsync_offset_hi << 4) + (_x).vsync_offset)
#define EDID_DETAILED_TIMING_VSYNC_PULSE_WIDTH(_x) \
	(((_x).vsync_pulse_width_hi << 4) + (_x).vsync_pulse_width)
	unsigned char himage_size;
	unsigned char vimage_size;
	unsigned char himage_size_hi: 4;
	unsigned char vimage_size_hi: 4;
#define EDID_DETAILED_TIMING_HIMAGE_SIZE(_x) \
	(((_x).himage_size_hi << 8) + (_x).himage_size)
#define EDID_DETAILED_TIMING_VIMAGE_SIZE(_x) \
	(((_x).vimage_size_hi << 8) + (_x).vimage_size)
	unsigned char hborder;
	unsigned char vborder;
	struct {
		unsigned char interlaced: 1;
		unsigned char stereo: 2;
		unsigned char digital_composite: 2;
		unsigned char variant: 2;
		unsigned char zero: 1;
	} flags;
};

enum {
	edid_monitor_descriptor_serial = 0xff,
	edid_monitor_descriptor_ascii = 0xfe,
	edid_monitor_descriptor_range = 0xfd,
	edid_monitor_descriptor_name = 0xfc,
	edid_monitor_descriptio_mf_range = 0xf,
} edid_monitor_descriptor_types;

struct edid_monitor_descriptor {
	uint16_t zero_flag_1;
	unsigned char zero_flag_2;
	unsigned char type;
	unsigned char zero_flag_3;
	union {
		char string[13];
		struct {
			unsigned char vertical_min;
			unsigned char vertical_max;
			unsigned char horizontal_min;
			unsigned char horizontal_max;
			unsigned char pixel_clock_max;
			unsigned char gtf_data[8];
		} range_data;
	} data;
};

struct edid_info_t {
	uint8_t  header[8];
	uint8_t  manufacturer_name[2];
	// First byte:  0-1 - second char LSBs, 2-6 - first char, 7 - zero
	// Second byte: 0-4 - third char, 7-5 - second char MSBs
	// chars is 5-bits
	uint8_t  product_code[2]; // LSB first
	uint32_t serial_number;
	unsigned char week;
	unsigned char year;
	unsigned char version;
	unsigned char revision;
	struct {
		unsigned char digital: 1;
		unsigned char separate_sync: 1;
		unsigned char composite_sync: 1;
		unsigned char sync_on_green: 1;
		unsigned char unused: 2;
		unsigned char voltage_level: 2;
	} video_input_definition;
	unsigned char max_size_horizontal;
	unsigned char max_size_vertical;
	unsigned char gamma;
	struct {
		unsigned char unused1: 3;
		unsigned char rgb: 1;
		unsigned char unused2: 1;
		unsigned char active_off: 1;
		unsigned char suspend: 1;
		unsigned char standby: 1;
	} feature_support;
	unsigned char color_characteristics[10];
	struct {
		unsigned char timing_720x400_70: 1;
		unsigned char timing_720x400_88: 1;
		unsigned char timing_640x480_60: 1;
		unsigned char timing_640x480_67: 1;
		unsigned char timing_640x480_72: 1;
		unsigned char timing_640x480_75: 1;
		unsigned char timing_800x600_56: 1;
		unsigned char timing_800x600_60: 1;
		unsigned char timing_800x600_72: 1;
		unsigned char timing_800x600_75: 1;
		unsigned char timing_832x624_75: 1;
		unsigned char timing_1024x768_87i: 1;
		unsigned char timing_1024x768_60: 1;
		unsigned char timing_1024x768_70: 1;
		unsigned char timing_1024x768_75: 1;
		unsigned char timing_1280x1024_75: 1;
	} established_timings;
	struct {
		unsigned char timing_1152x870_75: 1;
		unsigned char reserved: 7;
	} manufacturer_timings;
	struct {
		uint16_t xresolution: 8; // Horizontal resolution.  Multiply by 8, then add 248 for actual value
		uint16_t vfreq: 6;       // Vertical frequency. Add 60 to get actual value.
		uint16_t aspect: 2;      // 00=16:10, 01=4:3, 10=5:4, 11=16:9 (00=1:1 prior to v1.3)
	} standard_timing[8];
	union {
#if 0
		unsigned char detailed_timing[72];
#else
		struct edid_detailed_timing detailed_timing[4];
#endif
		struct edid_monitor_descriptor monitor_descriptor[4];
	} monitor_details;
	unsigned char extension_flag;
	unsigned char checksum;
};

#endif
